﻿using System.Windows.Forms;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Diagnostics;

namespace Shape_Mapper
{
    public enum GridMode { EditShape, SelectShape };

    /// <summary>
    /// Always work with the client size as we are using a panel and a panel can have
    ///  a border making the clientsize different from the size
    /// </summary>
    class Grid : Panel
    {
        private Pen _linePen = new Pen(Color.FromArgb(90, 134, 200), 1f);
        private BufferedGraphics _bufGFX;
        private BufferedGraphicsContext _bufContext = BufferedGraphicsManager.Current;
        private Point _nearestGridLock = new Point(-1, -1); // -1 -1 to indicate no nearest
        private Point _lastPointInShape;
        private Shape _shape = new Shape();
        private List<ReferenceLine> _refLineList = new List<ReferenceLine>();
        private int _gridDimension = 5;
        private int _vertexDim = 10;

        public event EventHandler ShapeChanged;

        public Grid()
        {
            _linePen.LineJoin = LineJoin.Round;
            _linePen.EndCap = LineCap.Round;
            _linePen.StartCap = LineCap.Round;

            this.BackColor = Color.FromArgb(69, 110, 180);
            this.BorderStyle = BorderStyle.FixedSingle;
            UpdateGraphicsBuffer();

            this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.UserPaint |
                ControlStyles.ResizeRedraw | ControlStyles.OptimizedDoubleBuffer | ControlStyles.Selectable, true);
        }

        public void Clear()
        {
            _shape.PlotList.Clear();
            _refLineList.Clear();
            OnShapeChanged();
            this.Invalidate();
        }

        protected void OnShapeChanged()
        {
            if (ShapeChanged != null)
            {
                ShapeChanged(this, EventArgs.Empty);
            }
        }

        private void UpdateGraphicsBuffer()
        {
            if (this.Width > 0 && this.Height > 0)
            {
                _bufContext.MaximumBuffer = new Size(this.ClientSize.Width + 1, this.ClientSize.Height + 1);
                _bufGFX = _bufContext.Allocate(this.CreateGraphics(), this.ClientRectangle);
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            if (_bufGFX != null)
            {
                _bufGFX.Graphics.Clear(this.BackColor);
                Render();
                _bufGFX.Render(e.Graphics);
            }
        }

        private void Render()
        {
            if (this.BackgroundImage != null)
            {
                _bufGFX.Graphics.DrawImage(this.BackgroundImage, this.ClientRectangle);
            }

            // Draw grid lines
            float heightSection = this.ClientSize.Height / (float)_gridDimension;
            float widthSection = this.ClientSize.Width / (float)_gridDimension;

            for (int i = 1; i < _gridDimension; i++)
            {
                _bufGFX.Graphics.DrawLine(_linePen, new PointF(0, heightSection * i), new PointF(this.ClientSize.Width, heightSection * i));
                _bufGFX.Graphics.DrawLine(_linePen, new PointF(widthSection * i, 0), new PointF(widthSection * i, this.ClientSize.Height));
            }

            {
                // Draw vertex (closest vertex to cursor)
                if (_nearestGridLock.X != -1)
                {
                    float x = _nearestGridLock.X - _vertexDim / 2f; // center it
                    float y = _nearestGridLock.Y - _vertexDim / 2f;
                    _bufGFX.Graphics.FillRectangle(Brushes.White, x, y, _vertexDim, _vertexDim);
                }
            }

            // Draw shape
            if (_shape.PlotList.Count > 1)
            {
                _shape.DrawShapeOutline(this);
            }

            _shape.DrawShapeNodes(this);

            // Draw reference lines
            foreach (ReferenceLine line in _refLineList)
            {
                if (line.Complete) line.DrawLine(this);
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            Point lastPos = GetNearestGridPos(e.Location);
            bool changed = (_nearestGridLock != lastPos && lastPos.X != -1);


            if (changed)
            {
                // Slightly increase update region to compensation for anti-aliasing
                Rectangle oldVertexRext = new Rectangle
                    (_nearestGridLock.X - _vertexDim / 2 - 1, _nearestGridLock.Y - _vertexDim / 2 - 1, _vertexDim + 2, _vertexDim + 2);

                Rectangle newVertexRext = new Rectangle
                    (lastPos.X - _vertexDim / 2 - 1, lastPos.Y - _vertexDim / 2 - 1, _vertexDim + 2, _vertexDim + 2);

                Rectangle updateRegion = Rectangle.Union(oldVertexRext, newVertexRext);

                _nearestGridLock = lastPos;

                this.Invalidate(updateRegion);
                Debug.WriteLine(updateRegion.ToString());
            }
        }

        /// <summary>
        /// Gets the location of the nearest intersecting point of the grid
        /// </summary>
        public Point GetNearestGridPos(Point pos)
        {
            // Set nearest grid lock
            float heightSection = (float)this.ClientSize.Height / _gridDimension;
            float widthSection = (float)this.ClientSize.Width / _gridDimension;
            int closest = int.MaxValue;
            Point lastPos = new Point(-1, -1);

            for (int i = 0; i <= _gridDimension; i++)
            {
                for (int i2 = 0; i2 <= _gridDimension; i2++)
                {
                    int x = (int)(i * widthSection + 0.5);
                    int y = (int)(i2 * heightSection + 0.5);
                    Point gridPos = new Point(x, y);
                    int xDiff = Math.Abs(gridPos.X - pos.X);
                    int yDiff = Math.Abs(gridPos.Y - pos.Y);

                    if (xDiff + yDiff < closest)
                    {
                        closest = xDiff + yDiff;
                        lastPos = gridPos;
                    }
                }
            }

            return lastPos;
        }

        public void DeleteNearestRefLine()
        {
            for (int i = 0; i < _refLineList.Count; i++)
            {
                if (_refLineList[i].EndPlot.GetLocation(this) == _nearestGridLock ||
                    _refLineList[i].StartPlot.GetLocation(this) == _nearestGridLock)
                {
                    _refLineList.Remove(_refLineList[i]);
                    this.Invalidate();
                }
            }
        }

        protected override void OnMouseClick(MouseEventArgs e)
        {
            base.OnMouseClick(e);

            if (e.Button == MouseButtons.Right)
            {
                // If last line is complete, then add new line
                if (_refLineList.Count == 0 || (_refLineList.Count > 0 && _refLineList[_refLineList.Count - 1].Complete))
                {
                    _refLineList.Add(new ReferenceLine());
                    _refLineList[_refLineList.Count - 1].StartPlot = new Plot(LocationToGridPoint(e.Location));
                }
                else
                {
                    _refLineList[_refLineList.Count - 1].EndPlot = new Plot(LocationToGridPoint(e.Location));
                }

                this.Invalidate();
            }
            else if (e.Button == MouseButtons.Left)
            {
                // Remove vertex if vertex is clicked
                if (_nearestGridLock == _lastPointInShape && _shape.PlotList.Count > 0)
                {
                    _shape.PlotList.RemoveAt(_shape.PlotList.Count - 1);

                    if (_shape.PlotList.Count > 0)
                    {
                        _lastPointInShape = _shape.PlotList[_shape.PlotList.Count - 1].GetLocation(this);
                    }
                    else
                    {
                        _lastPointInShape = new Point(-1, -1);
                    }

                    OnShapeChanged();
                    this.Invalidate();
                }
                else if (_lastPointInShape != _nearestGridLock && !IsFirstPointHovered)
                {
                    _lastPointInShape = _nearestGridLock;
                    _shape.PlotList.Add(new Plot(LocationToGridPoint(_nearestGridLock)));
                    OnShapeChanged();
                    this.Invalidate();
                }
            }
        }

        private Point LocationToGridPoint(Point location)
        {
            float heightSection = (float)this.ClientSize.Height / _gridDimension;
            float widthSection = (float)this.ClientSize.Width / _gridDimension;

            int x = (int)(location.X / widthSection + 0.5);
            int y = (int)(location.Y / heightSection + 0.5);
            return new Point(x, y);
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            UpdateGraphicsBuffer();
        }

        private bool IsFirstPointHovered
        {
            get
            {
                if (_shape.PlotList.Count > 0)
                {
                    return _shape.PlotList[0].GetLocation(this) == _nearestGridLock;
                }

                return false;
            }
        }

        /// <summary>
        /// Gets or sets how many grid cell across and vertical there are
        /// </summary>
        public int GridDimension
        {
            get { return _gridDimension; }
            set
            {
                if (value != _gridDimension) Clear(); // Clear if grid is changing
                _gridDimension = value;
                this.Invalidate();
            }
        }

        public Shape Shape
        {
            get { return _shape; }
        }

        public int VertexDimension
        {
            get { return _vertexDim; }
            set
            {
                _vertexDim = value;
                this.Invalidate();
            }
        }

        public Graphics BackBuffer
        {
            get { return _bufGFX.Graphics; }
        }
    }
}
